### ivoryos/optimizers/nimo_optimizer.py
import glob
import itertools
import os

import pandas as pd
from pandas import DataFrame

from ivoryos.optimizer.base_optimizer import OptimizerBase


class NIMOOptimizer(OptimizerBase):
    def __init__(self, experiment_name:str, parameter_space: list, objective_config: list, optimizer_config: dict,
                 parameter_constraints:list=None, datapath:str=None, additional_params:dict=None):
        """
        :param experiment_name: arbitrary name
        :param parameter_space: list of parameter names
        [
            {"name": "param_1", "type": "range", "bounds": [1.0, 2.0], "value_type": "float"},
            {"name": "param_2", "type": "choice", "bounds": ["a", "b", "c"], "value_type": "str"},
            {"name": "param_3", "type": "range", "bounds": [0 10], "value_type": "int"},
        ]
        :param objective_config: objective configuration
        [
            {"name": "obj_1", "minimize": True, "weight": 1},
            {"name": "obj_2", "minimize": False, "weight": 1}
        ]
        :param optimizer_config: optimizer configuration
        {
            "step_1": {"model": "RE", "num_samples": 10},
            "step_2": {"model": "PDC"}
        }
        """
        self.current_step = 0
        self.experiment_name = experiment_name
        self.parameter_space = parameter_space
        self.objective_config = objective_config
        self.optimizer_config = optimizer_config

        super().__init__(experiment_name, parameter_space, objective_config, optimizer_config, parameter_constraints, datapath, additional_params)

        os.makedirs(os.path.join(self.datapath, "nimo_data"), exist_ok=True)

        step_1 = optimizer_config.get("step_1", {})
        step_2 = optimizer_config.get("step_2", {})
        self.step_1_generator = step_1.get("model", "RE")
        self.step_1_batch_num = step_1.get("num_samples", 1)
        self.step_2_generator = step_2.get("model", "PDC")
        self.candidates = os.path.join(self.datapath, "nimo_data", f"{self.experiment_name}_candidates.csv")
        self.proposals = os.path.join(self.datapath, "nimo_data", f"{self.experiment_name}_proposals.csv")
        self.n_objectives = len(self.objective_config)
        self._create_candidates_csv()


    def _create_candidates_csv(self):
        """create candidates csv file for nimo input"""
        param_names = [p["name"] for p in self.parameter_space]

        param_values = []
        for p in self.parameter_space:
            if p["type"] == "choice" and isinstance(p["bounds"], list):
                param_values.append(p["bounds"])
            elif p["type"] == "range" and len(p["bounds"]) == 3:
                values = self._create_discrete_search_space(range_with_step=p["bounds"],value_type=p["value_type"])
                param_values.append(values)
            else:
                raise ValueError(f"Unsupported parameter format: {p}")

        # Generate all possible combinations
        combos = list(itertools.product(*param_values))

        # Create a DataFrame with parameter columns
        df = pd.DataFrame(combos, columns=param_names)
        # Add empty objective columns
        for obj in self.objective_config:
            df[obj["name"]] = ""

        # Save to CSV
        df.to_csv(self.candidates, index=False)


    def suggest(self, n=1):
        """suggest n candidates for next batch of trials"""
        import nimo

        method = self.step_1_generator if self.current_step < self.step_1_batch_num else self.step_2_generator

        nimo.selection(method = method,
                       input_file = self.candidates,
                       output_file = self.proposals,
                       num_objectives = self.n_objectives,
                       num_proposals = n,
                       **self.additional_params
                       )
        self.current_step += 1
        # Read proposals from CSV file
        proposals_df = pd.read_csv(self.proposals)
        # Get parameter names
        param_names = [p["name"] for p in self.parameter_space]
        # Convert proposals to list of parameter dictionaries
        proposals = []
        for _, row in proposals_df.iterrows():
            proposal = {name: row[name] for name in param_names}
            proposals.append(proposal)
        return proposals

    def _convert_observation_to_list(self, obs: dict) -> list:
        obj_names = [o["name"] for o in self.objective_config]
        return [obs.get(name, None) for name in obj_names]

    def observe(self, results: list):
        """
        observe single output, nimo obj input is [1,2,3] or [[1, 2], [1, 2], [1, 2]] for MO
        :param results: [{"objective_name": "value"}, {"objective_name": "value"}]]
        """
        import nimo
        nimo_objective_values = [self._convert_observation_to_list(result) for result in results]
        nimo.output_update(input_file=self.proposals,
                           output_file=self.candidates,
                           num_objectives=self.n_objectives,
                           objective_values=nimo_objective_values)

    def append_existing_data(self, existing_data: DataFrame, file_path: str = None):
        """
        append existing data to the candidates csv file for nimo input
        """
        import nimo
        num_objectives = len(self.objective_config)
        if file_path is None:
            return
        nimo.insert_objectives(input_file=file_path,
                               output_file=self.candidates,
                               num_objectives=num_objectives,
                               ndigits=2)

    def get_plots(self, plot_type):
        """requests phase diagram plot from nimo"""
        import nimo
        nimo.visualization.plot_phase_diagram.plot(input_file=self.candidates,
                                                   fig_folder=os.path.join(self.datapath, "nimo_data"))
        files = sorted(glob.glob(os.path.join(os.path.join(self.datapath, "nimo_data"), "phase_diagram_*.png")))
        if not files:
            return None
        return files[-1]

    @staticmethod
    def get_schema():
        return {
            "parameter_types": ["choice", "range"],
            "multiple_objectives": True,
            "supports_continuous": False,
            "supports_constraints": False,
            "optimizer_config": {
                "step_1": {"model": ["RE", "ES"], "num_samples": 5},
                "step_2": {"model": ["PHYSBO", "PDC", "BLOX", "PTR", "SLESA", "BOMP", "COMBI"]}
            },
            "additional_field":{
                "re_seed": {"type": "int", "required": False},
                "ptr_ranges": {"type": "list[float]", "required": False},

                "slesa_beta_max": {"type": "float", "required": False},
                "slesa_beta_num": {"type": "int", "required": False},

                "physbo_score": {"type": "choice", "options": ["Default", "EI", "PI", "TS", "HVPI", "EHVI", "TS"]},

                "pdc_estimation": {"type": "choice", "options": ["Default", "LP", "LS"]},
                "pdc_sampling": {"type": "choice", "options": ["Default", "LC", "MS", "EA"]},

                "process_X": {"type": "list[float]", "required": False},

                "combi_ranges": {"type": "list[float]", "required": False},
                "spread_elements": {"type": "list[int]", "required": False},

                "output_res": {"type": "choice", "options": ["Default", "True", "False"]},
                "training_res": {"type": "choice", "options": ["Default", "True", "False"]},
            }
        }




if __name__ == "__main__":
    parameter_space = [
        {"name": "silica", "type": "choice", "bounds": [100], "value_type": "float"},
        {"name": "water", "type": "range", "bounds": [500, 900, 50], "value_type": "float"},
        {"name": "PVA", "type": "choice", "bounds": [0, 0.005, 0.0075, 0.01, 0.05, 0.075, 0.1], "value_type": "float"},
        {"name": "SDS", "type": "choice", "bounds": [0], "value_type": "float"},
        {"name": "DTAB", "type": "choice", "bounds": [0, 0.005, 0.0075, 0.01, 0.05, 0.075, 0.1], "value_type": "float"},
        {"name": "PVP", "type": "choice", "bounds": [0], "value_type": "float"},
    ]
    objective_config = [
        {"name": "objective", "minimize": False, "weight": 1},

    ]
    optimizer_config = {
        "step_1": {"model": "RE", "num_samples": 10},
        "step_2": {"model": "PDC"}
    }

    nimo_optimizer = NIMOOptimizer(experiment_name="example_experiment", optimizer_config=optimizer_config, parameter_space=parameter_space, objective_config=objective_config)
    nimo_optimizer.suggest(n=1)
    nimo_optimizer.observe(
        results=[{"objective": 1.0}]
    )


import os
import uuid

import bcrypt
from flask import Flask, session, g, redirect, url_for
from flask_login import AnonymousUserMixin

from ivoryos.utils import utils
from ivoryos.utils.db_models import db, User
from ivoryos.routes.auth.auth import auth, login_manager
from ivoryos.routes.control.control import control
from ivoryos.routes.data.data import data
from ivoryos.routes.library.library import library
from ivoryos.routes.design.design import design
from ivoryos.routes.execute.execute import execute
from ivoryos.socket_handlers import socketio
from ivoryos.routes.main.main import main
from ivoryos.version import __version__ as ivoryos_version
from sqlalchemy import inspect, text

url_prefix = os.getenv('URL_PREFIX', "/ivoryos")
app = Flask(__name__, static_url_path=f'{url_prefix}/static', static_folder='static')
app.register_blueprint(main, url_prefix=url_prefix)
app.register_blueprint(auth, url_prefix=f'{url_prefix}/{auth.name}')
app.register_blueprint(library, url_prefix=f'{url_prefix}/{library.name}')
app.register_blueprint(control, url_prefix=f'{url_prefix}/instruments')
app.register_blueprint(design, url_prefix=f'{url_prefix}')
app.register_blueprint(execute, url_prefix=f'{url_prefix}')
app.register_blueprint(data, url_prefix=f'{url_prefix}')
# app.register_blueprint(api, url_prefix=f'{url_prefix}/{api.name}')

def reset_old_schema(engine, db_dir):
    inspector = inspect(engine)
    tables = inspector.get_table_names()


    # Check if old tables exist (no workflow_phases table)
    has_workflow_phase = 'workflow_phases' in tables
    old_workflow_run = 'workflow_runs' in tables
    old_workflow_step = 'workflow_steps' in tables

    # v1.3.4 only delete and backup when there is runs but no phases
    if not has_workflow_phase and old_workflow_run:
        print("⚠️ Old workflow database detected! All previous workflows have been reset to support the new schema.")
        # Backup old DB
        db_path = os.path.join(db_dir, "ivoryos.db")
        if os.path.exists(db_path):
            # os.makedirs(backup_dir, exist_ok=True)
            from datetime import datetime
            import shutil
            ts = datetime.now().strftime("%Y%m%d_%H%M%S")
            backup_path = os.path.join(db_dir, f"ivoryos_backup_{ts}.db")
            shutil.copy(db_path, backup_path)
            print(f"Backup created at {backup_path}")
        with engine.begin() as conn:
            # Drop old tables
            if old_workflow_step:
                conn.execute(text("DROP TABLE IF EXISTS workflow_steps"))
            if old_workflow_run:
                conn.execute(text("DROP TABLE IF EXISTS workflow_runs"))
    with engine.begin() as conn:
        try:
            conn.execute(
                text("ALTER TABLE user ADD COLUMN settings TEXT")
            )
        except Exception:
            pass
        try:
            conn.execute(text("ALTER TABLE script ADD COLUMN description TEXT"))
        except Exception:
            pass
    # Recreate new schema
    db.create_all()  # creates workflow_runs, workflow_phases, workflow_steps


def create_admin():
    """
    Create an admin user with username 'admin' and password 'admin' if it doesn't exist.
    """
    with app.app_context():
        admin_user = User.query.filter_by(username='admin').first()
        if not admin_user:
            print("Creating default admin user...")
            admin_user = User(
                username='admin',
                password=bcrypt.hashpw("admin".encode('utf-8'), bcrypt.gensalt()),
            )
            db.session.add(admin_user)
            db.session.commit()
        else:
            print("Admin user already exists.")


def create_app(config_class=None):
    """
    create app, init database
    """

    app.config.from_object(config_class or 'config.get_config()')
    os.makedirs(app.config["OUTPUT_FOLDER"], exist_ok=True)
    # Initialize extensions
    socketio.init_app(app, cors_allowed_origins="*", cookie=None)
    login_manager.init_app(app)
    login_manager.login_view = "auth.login"
    db.init_app(app)

    # Create database tables
    with app.app_context():
        # db.create_all()
        reset_old_schema(db.engine, app.config['OUTPUT_FOLDER'])
        create_admin()

    # Additional setup
    utils.create_gui_dir(app.config['OUTPUT_FOLDER'])

    # logger_list = app.config["LOGGERS"]
    logger_path = os.path.join(app.config["OUTPUT_FOLDER"], app.config["LOGGERS_PATH"])
    logger = utils.start_logger(socketio, 'gui_logger', logger_path)

    @app.before_request
    def before_request():
        """
        Called before

        """
        g.logger = logger
        g.socketio = socketio
        session.permanent = False
        # DEMO_MODE: Simulate logged-in user per session
        if app.config.get("DEMO_MODE", False):
            if "demo_user_id" not in session:
                session["demo_user_id"] = f"demo_{str(uuid.uuid4())[:8]}"

            class SessionDemoUser(AnonymousUserMixin):
                @property
                def is_authenticated(self):
                    return True

                def get_id(self):
                    return session.get("demo_user_id")

            login_manager.anonymous_user = SessionDemoUser



    @app.route('/')
    def redirect_to_prefix():
        return redirect(url_for('main.index', version=ivoryos_version))  # Assuming 'index' is a route in your blueprint

    @app.template_filter('format_name')
    def format_name(name):
        name = name.split(".")[-1]
        text = ' '.join(word for word in name.split('_'))
        return text.capitalize()

    # app.config.setdefault("DEMO_MODE", False)
    return app


import os

from flask import Blueprint, render_template, current_app, request, url_for
from flask_login import login_required, current_user
from werkzeug.utils import secure_filename, redirect
from ivoryos.utils.db_models import db
from ivoryos.version import __version__ as ivoryos_version

main = Blueprint('main', __name__, template_folder='templates')

@main.route("/")
@login_required
def index():
    """
    .. :quickref: Home page; ivoryos home page

    Home page for all available routes

    .. http:get:: /

    """
    off_line = current_app.config["OFF_LINE"]
    return render_template('home.html', off_line=off_line, version=ivoryos_version)


@main.route("/help")
def help_info():
    """
    .. :quickref: Help page; ivoryos info page

    static information page

    .. http:get:: /help

    """
    sample_deck = """
    from vapourtec.sf10 import SF10

    # connect SF10 pump
    sf10 = SF10(device_port="com7")

    # start ivoryOS
    from ivoryos.app import ivoryos
    ivoryos(__name__)
    """
    return render_template('help.html', sample_deck=sample_deck)


@main.route('/customize-logo', methods=['POST'])
@login_required
def customize_logo():
    if request.method == 'POST':
        file = request.files.get('logo')
        mode = request.form.get('mode')

        if file and file.filename != '':
            filename = secure_filename(file.filename)

            USER_LOGO_DIR = os.path.join(current_app.static_folder, "user_logos")
            os.makedirs(USER_LOGO_DIR, exist_ok=True)
            filepath = os.path.join(USER_LOGO_DIR, filename)
            file.save(filepath)

            # Save to database
            current_user.settings = {"logo_filename": filename, "logo_mode": mode}
            # current_user.logo_mode = mode
            db.session.commit()

        return redirect(url_for('main.index'))




from flask_login import current_user


def test_home_page_authenticated(auth, app):
    """
    GIVEN an authenticated user (using the 'auth' fixture)
    WHEN the home page is accessed
    THEN check that they see the main application page
    """
    with auth.application.test_request_context('/ivoryos/'):
        # Manually trigger the before_request functions that Flask-Login uses
        app.preprocess_request()

        # Assert that the `current_user` proxy is now populated and authenticated
        assert current_user.is_authenticated
        assert current_user.username == 'testuser'

def test_help_page(client):
    """
    GIVEN an unauthenticated user
    WHEN they access the help page
    THEN check that the page loads successfully and contains documentation content
    """
    response = client.get('/ivoryos/help')
    assert response.status_code == 200
    assert b'Documentations' in response.data

def test_prefix_redirect(auth):
    """
    GIVEN an authenticated user (using the 'auth' fixture)
    WHEN the home page is accessed without prefix
    THEN check that they see the main application page
    """
    response = auth.get('/', follow_redirects=True)
    assert response.status_code == 200


### Directory: ivoryos/optimizers/baybe_optimizer.py
from typing import Dict

from pandas import DataFrame

from ivoryos.utils.utils import install_and_import
from ivoryos.optimizer.base_optimizer import OptimizerBase

class BaybeOptimizer(OptimizerBase):
    def __init__(self, experiment_name, parameter_space, objective_config, optimizer_config,
                 parameter_constraints:list=None, datapath=None, additional_params:dict=None):
        try:
            from baybe import Campaign
        except ImportError:
            install_and_import("baybe")
            print("Please install Baybe with pip install baybe to before register BaybeOptimizer.")

        super().__init__(experiment_name, parameter_space, objective_config, optimizer_config, parameter_constraints, additional_params)
        self._trial_id = 0
        self._trials = {}

        self.experiment = Campaign(
            searchspace=self._convert_parameter_to_searchspace(parameter_space),
            objective=self._convert_objective_to_baybe_format(objective_config),
            recommender=self._convert_recommender_to_baybe_format(optimizer_config),
        )


    def suggest(self, n=1):
        # self.df = self.experiment.recommend(batch_size=n)
        return self.experiment.recommend(batch_size=n).to_dict(orient="records")

    def observe(self, results, index=None):
        """
        Observes the results of a trial and updates the experiment.
        :param results: A dictionary containing the results of the trial.
        :param index: The index of the trial in the DataFrame, if applicable.

        """
        df = DataFrame(results)
        self.experiment.add_measurements(df)

    def append_existing_data(self, existing_data: DataFrame, file_path: str = None):
        """
        Append existing data to the Ax experiment.
        :param existing_data: A dictionary containing existing data.
        """
        if existing_data.empty:
            return
        # parameter_names = [i.get("name") for i in self.parameter_space]
        # objective_names = [i.get("name") for i in self.objective_config]
        self.experiment.add_measurements(existing_data)
        # for name, value in existing_data.items():
        #     # First attach the trial and note the trial index
        #     parameters = {name: value for name in existing_data if name in parameter_names}
        #     trial_index = self.client.attach_trial(parameters=parameters)
        #     raw_data = {name: value for name in existing_data if name in objective_names}
        #     # Then complete the trial with the existing data
        #     self.client.complete_trial(trial_index=trial_index, raw_data=raw_data)


    def _convert_parameter_to_searchspace(self, parameter_space):
        """
        Converts the parameter space configuration to Baybe format.
        :param parameter_space: The parameter space configuration.
        [
            {"name": "param_1", "type": "range", "bounds": [1.0, 2.0], "value_type": "float"},
            {"name": "param_2", "type": "range", "bounds": [1.0, 2.0, 0.5], "value_type": "float"},

            {"name": "param_3", "type": "choice", "bounds": ["a", "b", "c"], "value_type": "str"},
            {"name": "param_4", "type": "range", "bounds": [0 10], "value_type": "int"},
            {"name": "param_5", "type": "substance", "bounds": ["methanol", "water", "toluene"], "value_type": "str"} #TODO
        ]
        :return: A list of Baybe parameters.
        """
        from baybe.parameters.categorical import CategoricalParameter
        from baybe.parameters.numerical import NumericalContinuousParameter, NumericalDiscreteParameter
        from baybe.searchspace import SearchSpace
        parameters = []
        for p in parameter_space:
            if p["type"] == "range":
                if len(p["bounds"]) == 3:
                    values = self._create_discrete_search_space(range_with_step=p["bounds"],value_type=p["value_type"])
                    parameters.append(NumericalDiscreteParameter(name=p["name"], values=values))
                elif p["value_type"] == "float":
                    parameters.append(NumericalContinuousParameter(name=p["name"], bounds=p["bounds"]))
                elif p["value_type"] == "int":
                    values = tuple([int(v) for v in range(p["bounds"][0], p["bounds"][1] + 1)])
                    parameters.append(NumericalDiscreteParameter(name=p["name"], values=values))

            elif p["type"] == "choice":
                if p["value_type"] == "str":
                    parameters.append(CategoricalParameter(name=p["name"], values=p["bounds"]))
                elif p["value_type"] in ["int", "float"]:
                    parameters.append(NumericalDiscreteParameter(name=p["name"], values=p["bounds"]))
        return SearchSpace.from_product(parameters)

    def _convert_objective_to_baybe_format(self, objective_config):
        """
        Converts the objective configuration to Baybe format.
        :param parameter_space: The parameter space configuration.
        [
            {"name": "obj_1", "minimize": True},
            {"name": "obj_2", "minimize": False}
        ]
        :return: A Baybe objective configuration.
        """
        from baybe.targets import NumericalTarget
        from baybe.objectives import SingleTargetObjective, DesirabilityObjective, ParetoObjective
        targets = []
        weights = []
        for obj in objective_config:
            obj_name = obj.get("name")
            minimize = obj.get("minimize", True)
            weight = obj.get("weight", 1)
            weights.append(weight)
            targets.append(NumericalTarget(name=obj_name, minimize=minimize))

        if len(targets) == 1:
            return SingleTargetObjective(target=targets[0])
        else:
            # Handle multiple objectives
            return ParetoObjective(targets=targets)


    def _convert_recommender_to_baybe_format(self, recommender_config):
        """
        Converts the recommender configuration to Baybe format.
        :param recommender_config: The recommender configuration.
        :return: A Baybe recommender configuration.
        """
        from baybe.recommenders import (
            BotorchRecommender,
            FPSRecommender,
            TwoPhaseMetaRecommender,
            RandomRecommender,
            NaiveHybridSpaceRecommender
        )
        step_1 = recommender_config.get("step_1", {})
        step_2 = recommender_config.get("step_2", {})
        step_1_recommender = step_1.get("model", "Random")
        step_2_recommender = step_2.get("model", "BOTorch")
        if step_1.get("model") == "Random":
            step_1_recommender = RandomRecommender()
        elif step_1.get("model") == "FPS":
            step_1_recommender = FPSRecommender()
        if step_2.get("model") == "Naive Hybrid Space":
            step_2_recommender = NaiveHybridSpaceRecommender()
        elif step_2.get("model") == "BOTorch":
            step_2_recommender = BotorchRecommender()
        return TwoPhaseMetaRecommender(
            initial_recommender=step_1_recommender,
            recommender=step_2_recommender
        )

    def get_plots(self, plot_type):
        return None

    @staticmethod
    def get_schema():
        """
        Returns a template for the optimizer configuration.
        """
        return {
            "parameter_types": ["range", "choice", "substance"],
            "multiple_objectives": True,
            "supports_continuous": True,
            "supports_constraints": False,
            "optimizer_config": {
                "step_1": {"model": ["Random", "FPS"], "num_samples": 10},
                "step_2": {"model": ["BOTorch", "Naive Hybrid Space"]}
            },
            "additional_field": {}
        }

if __name__ == "__main__":
    # Example usage
    baybe_optimizer = BaybeOptimizer(
        experiment_name="example_experiment",
        parameter_space=[
            {"name": "param_1", "type": "range", "bounds": [1.0, 2.0], "value_type": "float"},
            {"name": "param_2", "type": "choice", "bounds": ["a", "b", "c"], "value_type": "str"},
            {"name": "param_3", "type": "range", "bounds": [0, 10], "value_type": "int"}
        ],
        objective_config=[
            {"name": "obj_1", "minimize": True},
            {"name": "obj_2", "minimize": False}
        ],
        optimizer_config={
            "step_1": {"model": "Random", "num_samples": 10},
            "step_2": {"model": "BOTorch"}
        }
    )
    print(baybe_optimizer.suggest(5))